import contextlib
import typing
from types import TracebackType

from ozobot.common.sync import as_sync_context_manager
from ozobot.evo.driver import EvoDriver, get_driver
from ozobot.linefollower.api.handle import BaseHandle

from .core import Evo
from .sync import SyncEvo


class EvoHandle(BaseHandle):
    def __init__(self, *, address: str | None = None, id: str | None = None, name: str | None = None) -> None:
        super().__init__(address=address, id=id, name=name)
        self._exit_stack = contextlib.AsyncExitStack()

    async def __aenter__(self) -> Evo:
        Driver = get_driver()
        # the cast is essential for mypy to correctly infer `enter_async_context` return value
        driver_context = typing.cast(
            contextlib.AbstractAsyncContextManager[EvoDriver],
            Driver.open(address=self.address, id=self.id, name=self.name),
        )
        driver = await self._exit_stack.enter_async_context(driver_context)
        return Evo(driver)

    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)


class SyncEvoHandle(BaseHandle):
    def __init__(self, *, address: str | None = None, id: str | None = None, name: str | None = None) -> None:
        super().__init__(address=address, id=id, name=name)
        self._exit_stack = contextlib.ExitStack()

    def __enter__(self) -> SyncEvo:
        async_handle = EvoHandle(address=self.address, id=self.id, name=self.name)
        handle_context = self._exit_stack.enter_context(as_sync_context_manager(async_handle))
        return SyncEvo(handle_context)

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        self._exit_stack.__exit__(exc_type, exc_val, exc_tb)



from ozobot.evo.driver.shared import geometry
from ozobot.evo.webprotocol import types as webtypes
from ozobot.linefollower.datatypes import Sample
from ozobot.linefollower.driver.web import (
    LineFollowerWebDriver,
    Rpc,
    WebDataAccessReadWatch,
    WebMemoryRegions,
    rpctypes,
)
from ozobot.linefollower.driver.web.conversions import ir_message_from_web


class EvoWebMemoryRegions(WebMemoryRegions):
    def __init__(self, rpc: Rpc) -> None:
        super().__init__(rpc)

        self.ir_message_left_rear = WebDataAccessReadWatch(
            rpc,
            "irMessageLeftRear",
            response_type=rpctypes.ReadIrResponse,
            from_protocol=lambda m: Sample(ir_message_from_web(m), m.timestamp),
        )
        self.ir_message_right_rear = WebDataAccessReadWatch(
            rpc,
            "irMessageRightRear",
            response_type=rpctypes.ReadIrResponse,
            from_protocol=lambda m: Sample(ir_message_from_web(m), m.timestamp),
        )

        self.obstacle_left_rear = WebDataAccessReadWatch(
            rpc,
            "proximityLeftRear",
            response_type=rpctypes.IrProximityResponse,
            from_protocol=lambda m: Sample(m.value, m.timestamp),
        )
        self.obstacle_right_rear = WebDataAccessReadWatch(
            rpc,
            "proximityRightRear",
            response_type=rpctypes.IrProximityResponse,
            from_protocol=lambda m: Sample(m.value, m.timestamp),
        )

        self.button = WebDataAccessReadWatch(
            rpc,
            "button",
            response_type=webtypes.ButtonResponse,
            from_protocol=lambda m: Sample(m.press, m.timestamp),
        )
        self.charger = WebDataAccessReadWatch(
            rpc,
            "charger",
            response_type=webtypes.ChargerStateResponse,
            from_protocol=lambda m: Sample(m.state, m.timestamp),
        )

        self.geometry = geometry


class EvoWebDriver(LineFollowerWebDriver):
    @property
    def memory(self) -> EvoWebMemoryRegions:
        return self._evo_memory

    def __init__(self, name: str) -> None:
        super().__init__(name)
        self._evo_memory = EvoWebMemoryRegions(self._rpc)



import typing

from ozobot.evo.exceptions import OzobotDataTypeError, UnsupportedColorCodeNumberError
from ozobot.evo.protocol import Types
from ozobot.linefollower.datatypes import ColorCode, Direction, IRMessage, LEDMask, NamedColor


def _color_from_color_code_number(num: int) -> NamedColor:
    if num == 0:
        return NamedColor.BLACK
    elif num == 1:
        return NamedColor.RED
    elif num == 2:
        return NamedColor.GREEN
    elif num == 4:
        return NamedColor.BLUE

    raise UnsupportedColorCodeNumberError(num)


def color_code_from_protocol(color_code: Types.ColorCode) -> ColorCode:
    if not isinstance(color_code, Types.ColorCode):
        raise OzobotDataTypeError(Types.ColorCode, type(color_code))
    colors: list[NamedColor] = []
    num = color_code.code
    while num > 0:
        mod = num & 0b111
        num = num >> 3
        colors.append(_color_from_color_code_number(mod))
    return ColorCode(colors=tuple(colors))


def surface_color_from_protocol(surface_color: Types.SurfaceColor) -> NamedColor | None:
    if not isinstance(surface_color, Types.SurfaceColor):
        raise OzobotDataTypeError(Types.SurfaceColor, type(surface_color))

    match surface_color.color:
        case Types.SurfaceColorEnum.Black:
            return NamedColor.BLACK
        case Types.SurfaceColorEnum.Blue:
            return NamedColor.BLUE
        case Types.SurfaceColorEnum.Green:
            return NamedColor.GREEN
        case Types.SurfaceColorEnum.Red:
            return NamedColor.RED
        case Types.SurfaceColorEnum.White:
            return NamedColor.WHITE
        case Types.SurfaceColorEnum.Unknown:
            return None
        case _:
            typing.assert_never(surface_color.color)


def line_color_from_protocol(line_color: Types.LineColor) -> NamedColor | None:
    if not isinstance(line_color, Types.LineColor):
        raise OzobotDataTypeError(Types.LineColor, type(line_color))

    match line_color.color:
        case Types.LineColorEnum.Black:
            return NamedColor.BLACK
        case Types.LineColorEnum.Blue:
            return NamedColor.BLUE
        case Types.LineColorEnum.Green:
            return NamedColor.GREEN
        case Types.LineColorEnum.Red:
            return NamedColor.RED
        case Types.LineColorEnum.Unknown:
            return None
        case _:
            typing.assert_never(line_color.color)


def led_to_protocol(mask: LEDMask) -> Types.LEDsMask:
    if not isinstance(mask, LEDMask):
        raise OzobotDataTypeError(LEDMask, type(mask))

    protocol_mask: Types.LEDsMask = Types.LEDsMask(0)
    all_front_mask = (
        Types.LEDsMask.front_left
        | Types.LEDsMask.front_left_center
        | Types.LEDsMask.front_center
        | Types.LEDsMask.front_right_center
        | Types.LEDsMask.front_right
    )
    for led in mask:
        match led:
            case LEDMask.FRONT_LEFT:
                protocol_mask |= Types.LEDsMask.front_left
            case LEDMask.FRONT_LEFT_CENTER:
                protocol_mask |= Types.LEDsMask.front_left_center
            case LEDMask.FRONT_CENTER:
                protocol_mask |= Types.LEDsMask.front_center
            case LEDMask.FRONT_RIGHT_CENTER:
                protocol_mask |= Types.LEDsMask.front_right_center
            case LEDMask.FRONT_RIGHT:
                protocol_mask |= Types.LEDsMask.front_right
            case LEDMask.TOP:
                protocol_mask |= Types.LEDsMask.top
            case LEDMask.BACK:
                protocol_mask |= Types.LEDsMask.back
            case LEDMask.BUTTON:
                protocol_mask |= Types.LEDsMask.button
            # ALL_FRONT and ALL_ROBOT should not be needed because the iterator only goes through
            # single bit values, but let's make mypy happy
            case LEDMask.ALL_FRONT:
                protocol_mask |= all_front_mask
            case LEDMask.ALL_ROBOT:
                protocol_mask |= all_front_mask | Types.LEDsMask.top | Types.LEDsMask.back | Types.LEDsMask.button
            case _:
                typing.assert_never(led)

    return protocol_mask


def intersection_direction_to_protocol(direction: Direction) -> Types.IntersectionDirection:
    if not isinstance(direction, Direction):
        raise OzobotDataTypeError(Direction, type(direction))

    if not len(direction) == 1:
        raise ValueError("Direction attribute needs to define exactly one direction")

    match direction:
        case Direction.LEFT:
            return Types.IntersectionDirection.Left
        case Direction.RIGHT:
            return Types.IntersectionDirection.Right
        case Direction.STRAIGHT:
            return Types.IntersectionDirection.Straight
        case Direction.BACKWARD:
            return Types.IntersectionDirection.Backward
        case _:
            typing.assert_never(direction)


def intersection_bitmap_from_protocol(intersection_mask: Types.IntersectionBitmap) -> Direction:
    if not isinstance(intersection_mask, Types.IntersectionBitmap):
        raise OzobotDataTypeError(Types.IntersectionBitmap, type(intersection_mask))

    intersection = Direction(0)
    for dir in intersection_mask:
        match dir:
            case Types.IntersectionBitmap.Backward:
                intersection |= Direction.BACKWARD
            case Types.IntersectionBitmap.Straight:
                intersection |= Direction.STRAIGHT
            case Types.IntersectionBitmap.Left:
                intersection |= Direction.LEFT
            case Types.IntersectionBitmap.Right:
                intersection |= Direction.RIGHT
            case _:
                typing.assert_never(dir)

    return intersection


def ir_message_from_protocol(ir_message: Types.IRMessage) -> IRMessage:
    return IRMessage(
        message=ir_message.message,
        intensity=ir_message.intensity,
    )


def charger_state_from_protocol(
    charger_state: Types.ChargerState,
) -> typing.Literal["Connected", "Disconnected", "LowPower"]:
    match charger_state.state:
        case Types.ChargerStateEnum.Connected:
            return "Connected"
        case Types.ChargerStateEnum.Disconnected:
            return "Disconnected"
        case Types.ChargerStateEnum.LowPower:
            return "LowPower"
        case _:
            typing.assert_never(charger_state.state)



import sys
import typing

if typing.TYPE_CHECKING:
    from ozobot.evo.driver.native import EvoNativeDriver
    from ozobot.evo.driver.web import EvoWebDriver

    type EvoDriver = EvoWebDriver | EvoNativeDriver
else:
    type EvoDriver = typing.Any


def get_driver() -> type[EvoDriver]:
    # don't use if sys.platform directly, mypy will then only check the platform specific branch
    platform: str = sys.platform

    if platform == "emscripten":
        from ozobot.evo.driver.web import EvoWebDriver

        return EvoWebDriver
    else:
        raise NotImplementedError("Evo native driver is not yet implemented")
        from ozobot.evo.driver.native import EvoNativeDriver

        return EvoNativeDriver



from __future__ import annotations

import asyncio
import typing

from loguru import logger
from ozobot.evo.driver import EvoDriver
from ozobot.linefollower.api.core import LineFollower
from ozobot.linefollower.datatypes import IRMessage, Sample
from ozobot.linefollower.driver.interface import ReadableWatchableRegion, VirtualMemoryRegions


class EvoVirtualMemory(VirtualMemoryRegions, typing.Protocol):
    @property
    def obstacle_right_rear(self) -> ReadableWatchableRegion[Sample[float]]:
        """
        Right rear IR sensor's measured intensity reflected from obstacle.
        """

    @property
    def obstacle_left_rear(self) -> ReadableWatchableRegion[Sample[float]]:
        """
        Left rear IR sensor's measured intensity reflected from obstacle.
        """

    @property
    def ir_message_left_rear(self) -> ReadableWatchableRegion[Sample[IRMessage]]:
        """
        Message received by the left rear IR sensor.
        """

    @property
    def ir_message_right_rear(self) -> ReadableWatchableRegion[Sample[IRMessage]]:
        """
        Message received by the right rear IR sensor.
        """


# this enables verbose errors when memory region implementations do not
# match the interfaces
if typing.TYPE_CHECKING:  # type: ignore[name-defined]
    _vm: EvoVirtualMemory
    from ozobot.evo.driver.native import NativeMemoryRegions
    from ozobot.evo.driver.web import EvoWebMemoryRegions

    _vm = EvoWebMemoryRegions()  # type: ignore[call-arg]
    _vm = NativeMemoryRegions()  # type: ignore[call-arg]


class Evo(LineFollower):
    @property
    def data(self) -> EvoVirtualMemory:
        return self._evo_driver.memory

    def __init__(self, driver: EvoDriver) -> None:
        super().__init__(driver)
        self._evo_driver = driver
        self._set_velocity_overridden = asyncio.Condition()

    async def set_velocity(self, linear_mmps: float, angular_degps: float, duration_s: float) -> None:
        logger.debug("Setting velocity", linear=linear_mmps, angular=angular_degps, duration=duration_s)

        # If the velocity command is called while another is running, Evo does not send any notification about this. This is problem especially
        # when a `set_velocity` call is made while another `set_velocity` execution is being executed in a task. The new call overrides the previous one,
        # but the library does not get any notification about that and the task hangs.
        #
        # This workaround makes the function also finish on a condition triggered by calling `set_velocity` rising asyncio.CancellationError
        async with self._set_velocity_overridden:
            self._set_velocity_overridden.notify()

            coros = [
                self._driver.velocity(
                    linear_mmps,
                    angular_degps,
                    int(duration_s * 1000),
                ),
                self._set_velocity_overridden.wait(),
            ]

            async with asyncio.TaskGroup() as tg:
                tasks = [tg.create_task(c) for c in coros]
                _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

                # cancel pending task and propagate exception from the finished task
                for t in tasks:
                    if t.done():
                        await t
                    else:
                        t.cancel()



from ozobot.linefollower.api.data_access import DataReadConstant
from ozobot.linefollower.datatypes import RobotGeometry

geometry = DataReadConstant(
    lambda: RobotGeometry(
        ticks_per_mm=18.851,
        wheel_track_mm=23,
        wheel_diameter_mm=11.82,
        encoder_ticks_per_wheel_revolution=8 * 2 * 21 * 25 / 12,
        max_speed_limit_mmps=300,
    )
)



from ozobot.evo.api.core import Evo
from ozobot.evo.api.handle import EvoHandle, SyncEvoHandle
from ozobot.evo.api.sync import SyncEvo

__all__ = [
    "Evo",
    "EvoHandle",
    "SyncEvoHandle",
    "SyncEvo",
]



import typing

from ozobot.evo.exceptions import OzobotProtocolCommandError
from ozobot.evo.protocol import Types
from ozobot.linefollower.exceptions import LinefollowerFileNotFoundError


class _HasExecutionState(typing.Protocol):
    executionState: Types.ExecutionStateEnum


@typing.runtime_checkable
class _HasCallStatus(typing.Protocol):
    callStatus: Types.CallStatus


@typing.runtime_checkable
class _HasResult(typing.Protocol):
    result: Types.IOResult


def handle_response(function_name: str, response: _HasCallStatus | _HasResult) -> None:
    if isinstance(response, _HasCallStatus):
        if response.callStatus != Types.CallStatus.CallSuccess:
            raise OzobotProtocolCommandError(function_name, response.callStatus.name, description="call failed")
    elif isinstance(response, _HasResult):
        if response.result != Types.IOResult.Success:
            raise OzobotProtocolCommandError(function_name, response.result.name, description="call failed")


async def handle_events[T: _HasExecutionState](function_name: str, events: typing.AsyncIterator[T]) -> T:
    """Checks event execution state and returns an event confirming event success. Raises exception otherwise."""
    async for event in events:
        if event.executionState == Types.ExecutionStateEnum.FinishedNormal:
            return event

        if event.executionState == Types.ExecutionStateEnum.FileNotFound:
            raise LinefollowerFileNotFoundError()

        if event.executionState != Types.ExecutionStateEnum.Running:
            raise OzobotProtocolCommandError(
                function_name,
                event.executionState.name,
                description="failure execution state",
            )

    raise OzobotProtocolCommandError(function_name, "empty", description="no command end state received")



from ozobot.evo.api.core import Evo
from ozobot.linefollower.api.sync import SyncLineFollower, SyncMemoryRegions


class SyncEvoVirtualMemory(SyncMemoryRegions): ...


class SyncEvo(SyncLineFollower):
    @property
    def data(self) -> SyncEvoVirtualMemory:
        """
        Robot sensors.

        Contains virtual memory and sensor structures allowing a subset of read, write and watch methods.
        """

        return SyncEvoVirtualMemory(self._evo)

    def __init__(self, evo: Evo) -> None:
        self._evo = evo
        super().__init__(evo)



import typing

from ozobot.linefollower.driver.web import rpctypes


class ButtonResponse(rpctypes.BaseResponse):
    press: bool
    timestamp: int


class ChargerStateResponse(rpctypes.BaseResponse):
    state: typing.Literal["Connected", "Disconnected", "LowPower"]
    timestamp: int



from __future__ import annotations

import asyncio
import contextlib
import typing
import uuid
from dataclasses import dataclass

from loguru import logger
from ozobot.common.broadcast import BroadcastManager
from ozobot.evo.driver.responses import handle_response
from ozobot.evo.protocol import AsyncControl, PacketTypes, Types, VirtualMemory


class _Deserializable(typing.Protocol):
    @classmethod
    def deserialize(cls, data: bytes) -> typing.Self: ...


@dataclass(frozen=True, kw_only=True)
class _WatcherAllocation[T]:
    watcher_id: int
    region_id: int
    address: int
    size: int
    type: type[T]

    @property
    def end(self) -> int:
        return self.address + self.end


@dataclass(frozen=True, kw_only=True)
class _WatcherAllocationSummary:
    used_regions: list[int]
    num_used_bytes: int


class _SubscriptionInfo[T](typing.Protocol):
    address: int
    type: type[T]

    @property
    def size(self) -> int: ...


class WatcherSubscription[T: _Deserializable]:
    @property
    def last(self) -> T:
        return self._last_value

    def __init__(self, initial_value: T) -> None:
        self._broadcast = BroadcastManager[T]()
        self._last_value = initial_value

    @contextlib.asynccontextmanager
    async def watch(self) -> typing.AsyncIterator[typing.AsyncIterator[T]]:
        with self._broadcast.output() as output:

            async def _read() -> typing.AsyncIterator[T]:
                while True:
                    yield await output.get()

            yield _read()

    @classmethod
    @contextlib.asynccontextmanager
    async def run(
        cls,
        events: typing.AsyncIterator[PacketTypes.PacketEvent_WatcherDirty],
        allocations: list[_WatcherAllocation],
        allocation: _WatcherAllocation[T],
    ) -> typing.AsyncIterator[WatcherSubscription[T]]:
        async with asyncio.TaskGroup() as tg:
            parsed_events = cls._parse(events, allocations, allocation)
            initial_value = await anext(parsed_events)
            subs = WatcherSubscription(initial_value)
            process_task = tg.create_task(subs._process(parsed_events))

            try:
                yield subs
            finally:
                process_task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await process_task

    @classmethod
    async def _parse(
        cls,
        events: typing.AsyncIterator[PacketTypes.PacketEvent_WatcherDirty],
        allocations: list[_WatcherAllocation],
        allocation: _WatcherAllocation[T],
    ) -> typing.AsyncGenerator[T, None]:
        logger.debug("finding offset", allocation=allocation, allocations=allocations)
        watcher_regions = list(filter(lambda e: e.watcher_id == allocation.watcher_id, allocations))
        preceding_regions = list(filter(lambda e: e.region_id < allocation.region_id, watcher_regions))

        logger.debug("computing offset", regions=preceding_regions)
        expected_len = sum(a.size for a in watcher_regions)
        offset = sum(r.size for r in preceding_regions)
        logger.debug("got offset", offset=offset, size=allocation.size)

        async for event in events:
            data = bytes(event.data)

            if len(data) != expected_len:
                logger.warning("WatcherDirty event length wrong, skipping", expected=expected_len, actual=len(data))
            else:
                yield allocation.type.deserialize(data[offset : offset + allocation.size])

    async def _process(self, events: typing.AsyncIterator[T]) -> None:
        async for event in events:
            self._last_value = event
            await self._broadcast.broadcast(event)


class WatcherAllocator:
    def __init__(self, watcher_count: int, region_count: int, packet_size_max: int) -> None:
        self._watcher_count = watcher_count
        self._region_count = region_count
        self._packet_size_max = packet_size_max
        self._allocations: dict[int, list[_WatcherAllocation]] = {watcher_id: [] for watcher_id in range(watcher_count)}

    def allocate[T](self, size: int, address: int, _type: type[T]) -> _WatcherAllocation[T]:
        watcher_id, region_id = self._find_free_gap(size)
        allocation = _WatcherAllocation(
            watcher_id=watcher_id, region_id=region_id, size=size, address=address, type=_type
        )
        self._allocations[watcher_id].append(allocation)

        return allocation

    def _find_free_gap(self, size: int) -> tuple[int, int]:
        max_size = self._packet_size_max - PacketTypes.PacketEvent_WatcherDirty.expected_length(None)

        logger.debug("Searching for free gap", size=size, max_size=max_size, allocations=self._allocations)
        allocation_summary = {
            watcher_id: _WatcherAllocationSummary(
                used_regions=[a.region_id for a in self._allocations.get(watcher_id, [])],
                num_used_bytes=sum([a.size for a in self._allocations.get(watcher_id, [])]),
            )
            for watcher_id in range(self._watcher_count)
        }

        # filter out watchers with no regions left or not enough free packet size
        available_watchers = {
            watcher_id: summary
            for watcher_id, summary in allocation_summary.items()
            if len(summary.used_regions) < self._region_count and max_size - summary.num_used_bytes >= size
        }

        logger.debug("Available watchers", available_watchers=available_watchers)

        # find watcher with lowest free memory (smallest gap) to prevent fragmentation
        watcher_id, watcher_allocation = max(
            available_watchers.items(), default=(None, None), key=lambda e: e[1].num_used_bytes
        )
        if watcher_id is None or watcher_allocation is None:
            raise Exception("No free watcher found")

        # get the first unused
        unused_region_ids = set(range(self._region_count)) - set(watcher_allocation.used_regions)
        region_id = min(unused_region_ids)

        return watcher_id, region_id


class LineFollowerWatcher:
    def __init__(self, control: AsyncControl) -> None:
        self._control = control
        self._watcher_enabled = False

    # FIXME: the return type is not perfect, when/if https://github.com/python/typing/issues/1383 is resolved, fix this
    @contextlib.asynccontextmanager
    async def watch[T: _Deserializable](
        self, subscription_configs: tuple[_SubscriptionInfo[T], ...]
    ) -> typing.AsyncIterator[tuple[WatcherSubscription[T], ...]]:
        if self._watcher_enabled:
            raise Exception("Watcher already enabled")

        self._watcher_enabled = True

        async with contextlib.AsyncExitStack() as exit_stack:
            watcher_info = await self._read_watcher_info()
            allocator = WatcherAllocator(
                watcher_info.watcherCount, watcher_info.watcherRegionCount, self._control.packet_size_max
            )

            allocations = [allocator.allocate(sub.size, sub.address, sub.type) for sub in subscription_configs]
            watchers = [await exit_stack.enter_async_context(self._subscribe(allocation)) for allocation in allocations]
            watcher_ids = [allocation.watcher_id for allocation in allocations]

            async with self._enable_watchers(watcher_ids):
                subscriptions = [
                    await exit_stack.enter_async_context(WatcherSubscription.run(watcher, allocations, allocation))
                    for watcher, allocation in zip(watchers, allocations, strict=False)
                ]
                yield tuple(subscriptions)

            self._watcher_enabled = False

    @contextlib.asynccontextmanager
    async def _enable_watchers(self, watcher_ids: list[int]) -> typing.AsyncIterator[None]:
        logger.debug("Initializing watchers", watcher_ids=watcher_ids)
        flags = (
            Types.WatcherFlags.Enabled
            | Types.WatcherFlags.DisableWhenDisconnected
            | Types.WatcherFlags.SendInitialValue
        )
        for watcher_id in watcher_ids:
            async with self._control.WatcherSetup(watcher_id, flags, 40, 0) as (resp, _):
                handle_response("WatcherSetup", resp)

        try:
            yield
        finally:
            logger.debug("Deinitializing watchers", watcher_ids=watcher_ids)
            for watcher_id in watcher_ids:
                async with self._control.WatcherSetup(watcher_id, Types.WatcherFlags(0), 0, 0) as (resp, _):
                    handle_response("WatcherSetup", resp)

    @contextlib.asynccontextmanager
    async def _subscribe[T: _Deserializable](
        self, allocation: _WatcherAllocation[T]
    ) -> typing.AsyncIterator[typing.AsyncIterator[PacketTypes.PacketEvent_WatcherDirty]]:
        with logger.contextualize(watcher=uuid.uuid4()):
            logger.debug("Subscribing watcher", address=allocation.address, size=allocation.size)

            watcher_id, region_id = allocation.watcher_id, allocation.region_id
            logger.debug("Found watcher", watcher_id=watcher_id, region_id=region_id)
            try:
                flags = Types.WatcherRegionFlags(0)
                region_setup_rpc = self._control.WatcherRegionSetup(
                    watcher_id, region_id, allocation.address, allocation.size, flags=flags
                )
                async with region_setup_rpc as (reply, events):
                    yield events
            finally:
                logger.debug("Removing watcher")
                flags = Types.WatcherRegionFlags.DoNotSendInNotify | Types.WatcherRegionFlags.DoNotSetDirty
                async with self._control.WatcherRegionSetup(watcher_id, region_id, 0, 0, flags=flags) as (resp, _):
                    handle_response("WatcherRegionSetup", resp)

    async def _read_watcher_info(self) -> Types.WatcherInfo:
        metadata = VirtualMemory.watchersInfo
        async with self._control.MemRead(metadata.address, metadata.type.data_width) as (resp, _):
            handle_response("MemRead", resp)
        data = bytes(resp.data)
        return metadata.type.deserialize(data)



from __future__ import annotations

import contextlib
import math
import typing
from uuid import UUID

from ozobot.ble.connection import open_client
from ozobot.evo import conversions
from ozobot.evo.api.watchers import LineFollowerWatcher, WatcherSubscription
from ozobot.evo.driver.responses import handle_events, handle_response
from ozobot.evo.driver.shared import geometry
from ozobot.evo.protocol import AsyncControl, Types, VirtualMemory
from ozobot.linefollower.api.data_access import (
    DataWatcherProxy,
    EventWatcher,
    EventWatcherQueue,
    WatcherOutputContainer,
    WatcherOutputContainerRunner,
)
from ozobot.linefollower.datatypes import (
    Direction,
    LEDMask,
    Sample,
    SampleWithoutTimestamp,
)
from ozobot.linefollower.driver.interface import Deserializable, Serializable

_SERVICE_UUID = UUID("8903136c-5f13-4548-a885-c58779136801")
_CHARACTERISTIC_UUID = UUID("8903136c-5f13-4548-a885-c58779136802")


class _DeserializeAndSerializable(Deserializable, Serializable, typing.Protocol):
    pass


class _DeserializableWithTimestamp(Deserializable, typing.Protocol):
    timestamp: int


class MemoryProperty[T](typing.Protocol):
    address: int
    type: type[T]

    @property
    def size(self) -> int: ...


type _TWatchers = tuple[
    WatcherSubscription[Types.ColorCode],
    WatcherSubscription[Types.LineColor],
    WatcherSubscription[Types.SurfaceColor],
    WatcherSubscription[Types.IRProximity],
    WatcherSubscription[Types.IRMessage],
    WatcherSubscription[Types.IRMessage],
    WatcherSubscription[Types.IRMessage],
    WatcherSubscription[Types.IRMessage],
    WatcherSubscription[Types.Button],
    WatcherSubscription[Types.ChargerState],
]


async def _stop_execution(control: AsyncControl, *, request_id: int) -> None:
    async with control.StopExecution(request_id) as (resp, _):
        pass  # there's nothing to check in the StopExecution response


class NativeMemoryRegions:
    def __init__(self, control: AsyncControl, watchers: _TWatchers) -> None:
        color_code_watcher, line_color_watcher, surface_color_watcher, proximity_watcher = watchers[:4]
        ir_right_front_watcher, ir_left_front_watcher, ir_right_rear_watcher, ir_left_rear_watcher = watchers[4:8]
        button_watcher, charger_watcher = watchers[8:10]

        self.intersection_queue = EventWatcherQueue(SampleWithoutTimestamp(Direction(0)))
        self.intersection = EventWatcher(self.intersection_queue)

        self.line_following_speed = NativeDataAccessReadWrite(
            control,
            VirtualMemory.lineNavigationSpeed,
            lambda s8_24: float(s8_24) * 1000,
            lambda fl: Types.S8_24(fl / 1000),
        )
        self.color_code = NativeDataWatcher(
            control,
            VirtualMemory.colorCode,
            color_code_watcher,
            lambda c: SampleWithoutTimestamp(conversions.color_code_from_protocol(c)),
        )
        self.line_color = NativeDataWatcher(
            control,
            VirtualMemory.lineColor,
            line_color_watcher,
            lambda c: Sample(conversions.line_color_from_protocol(c), c.timestamp),
        )
        self.surface_color = NativeDataWatcher(
            control,
            VirtualMemory.surfaceColor,
            surface_color_watcher,
            lambda c: Sample(conversions.surface_color_from_protocol(c), c.timestamp),
        )

        proximity = NativeDataWatcher(control, VirtualMemory.irProximity, proximity_watcher, lambda d: d)

        self.obstacle_right_front = DataWatcherProxy(proximity, convert=lambda p: Sample(p.right_front, p.timestamp))
        self.obstacle_left_front = DataWatcherProxy(proximity, convert=lambda p: Sample(p.left_front, p.timestamp))
        self.obstacle_right_rear = DataWatcherProxy(proximity, convert=lambda p: Sample(p.right_rear, p.timestamp))
        self.obstacle_left_rear = DataWatcherProxy(proximity, convert=lambda p: Sample(p.left_rear, p.timestamp))

        self.ir_message_right_front = NativeDataWatcher(
            control,
            VirtualMemory.irMessageRightFront,
            ir_right_front_watcher,
            lambda m: Sample(conversions.ir_message_from_protocol(m), m.timestamp),
        )

        self.ir_message_left_front = NativeDataWatcher(
            control,
            VirtualMemory.irMessageLeftFront,
            ir_left_front_watcher,
            lambda m: Sample(conversions.ir_message_from_protocol(m), m.timestamp),
        )

        self.ir_message_right_rear = NativeDataWatcher(
            control,
            VirtualMemory.irMessageRightRear,
            ir_right_rear_watcher,
            lambda m: Sample(conversions.ir_message_from_protocol(m), m.timestamp),
        )

        self.ir_message_left_rear = NativeDataWatcher(
            control,
            VirtualMemory.irMessageLeftRear,
            ir_left_rear_watcher,
            lambda m: Sample(conversions.ir_message_from_protocol(m), m.timestamp),
        )

        self.button = NativeDataWatcher(
            control, VirtualMemory.button, button_watcher, lambda b: Sample(b.press, b.timestamp)
        )
        self.charger = NativeDataWatcher(
            control,
            VirtualMemory.chargerState,
            charger_watcher,
            lambda c: Sample(conversions.charger_state_from_protocol(c), c.timestamp),
        )

        self.geometry = geometry


class NativeDataAccessRead[T: Deserializable, U]:
    def __init__(
        self, control: AsyncControl, property: MemoryProperty[T], from_protocol: typing.Callable[[T], U]
    ) -> None:
        self._control = control
        self._property = property
        self._from_protocol = from_protocol

    async def read(self) -> U:
        async with self._control.MemRead(self._property.address, self._property.size) as (resp, _):
            handle_response("MemRead", resp)

        val = self._property.type.deserialize(bytes(resp.data))
        return self._from_protocol(val)


class NativeDataAccessReadWrite[T: _DeserializeAndSerializable, U](NativeDataAccessRead[T, U]):
    def __init__(
        self,
        control: AsyncControl,
        property: MemoryProperty[T],
        from_protocol: typing.Callable[[T], U],
        to_protocol: typing.Callable[[U], T],
    ) -> None:
        super().__init__(control, property, from_protocol)
        self._to_protocol = to_protocol

    async def write(self, data: U) -> None:
        raw_data = self._to_protocol(data).serialize()
        async with self._control.MemWrite(self._property.address, self._property.size, raw_data) as (resp, _):
            handle_response("MemWrite", resp)


class NativeDataWatcher[T: _DeserializableWithTimestamp, U]:
    def __init__(
        self,
        control: AsyncControl,
        property: MemoryProperty[T],
        watcher: WatcherSubscription[T],
        from_protocol: typing.Callable[[T], U],
    ) -> None:
        self._watcher = watcher
        self._reader = NativeDataAccessRead(control, property, from_protocol)
        self._from_protocol = from_protocol

    async def read(self) -> U:
        return await self._reader.read()

    @contextlib.asynccontextmanager
    async def watch(self) -> typing.AsyncIterator[WatcherOutputContainer[U]]:
        async def _convert_iterator(it: typing.AsyncIterator[T]) -> typing.AsyncGenerator[U, None]:
            async for data in it:
                yield self._from_protocol(data)

        async with WatcherOutputContainerRunner[U]() as container_runner:
            async with self._watcher.watch() as unbuffered_reader:
                await container_runner.start(_convert_iterator(unbuffered_reader))
                yield container_runner.output_container


@contextlib.asynccontextmanager
async def create_memory_regions_structure(control: AsyncControl) -> typing.AsyncIterator[NativeMemoryRegions]:
    config = (
        VirtualMemory.colorCode,
        VirtualMemory.lineColor,
        VirtualMemory.surfaceColor,
        VirtualMemory.irProximity,
        VirtualMemory.irMessageRightFront,
        VirtualMemory.irMessageLeftFront,
        VirtualMemory.irMessageRightRear,
        VirtualMemory.irMessageLeftRear,
        VirtualMemory.button,
        VirtualMemory.chargerState,
    )

    watcher = LineFollowerWatcher(control)

    # Until typing allows tuple mapping, we need to do the casts below. See: https://github.com/python/typing/issues/1383
    type TExpectedInput = tuple[MemoryProperty[Types.ColorCode | Types.LineColor | Types.SurfaceColor]]

    async with watcher.watch(typing.cast(TExpectedInput, config)) as subscriptions:
        yield NativeMemoryRegions(control, typing.cast(_TWatchers, subscriptions))


class EvoNativeDriver:
    def __init__(self, control: AsyncControl, memory: NativeMemoryRegions) -> None:
        self._control = control
        self.memory = memory

    @classmethod
    @contextlib.asynccontextmanager
    async def open(
        cls,
        address: str | None = None,
        id: str | None = None,
        name: str | None = None,
    ) -> typing.AsyncIterator[EvoNativeDriver]:
        async with open_client(address=address, id=id, name=name) as client:
            char = client.get_characteristic(_SERVICE_UUID, _CHARACTERISTIC_UUID)
            control = AsyncControl(char)
            await _stop_execution(control, request_id=0)

            async with create_memory_regions_structure(control) as memory:
                yield cls(control, memory)

    async def move(self, distance_mm: float, speed_mmps: float) -> None:
        request_id = self._control.get_next_request_id()
        async with self._control.MoveStraight(request_id, distance_mm / 1000, speed_mmps / 1000) as (
            resp,
            evts,
        ):
            async with self._cancellation(request_id=request_id):
                await handle_events("MoveStraight", evts)

    async def rotate(self, angle_deg: float, angular_speed_degps: float) -> None:
        request_id = self._control.get_next_request_id()
        async with self._control.Rotate(request_id, math.radians(angle_deg), math.radians(angular_speed_degps)) as (
            resp,
            evts,
        ):
            async with self._cancellation(request_id=request_id):
                await handle_events("Rotate", evts)

    async def velocity(self, linear_mmps: float, angular_degps: float, duration_ms: int) -> None:
        request_id = self._control.get_next_request_id()
        async with self._control.Velocity(request_id, linear_mmps / 1000, math.radians(angular_degps), duration_ms) as (
            resp,
            evts,
        ):
            async with self._cancellation(request_id=request_id):
                await handle_events("Velocity", evts)

    async def play_tone(self, frequency_hz: int, duration_ms: int, volume: int) -> None:
        request_id = self._control.get_next_request_id()
        async with self._control.PlayTone(request_id, frequency_hz, duration_ms, volume) as (
            resp,
            evts,
        ):
            async with self._cancellation(request_id=request_id):
                await handle_events("PlayTone", evts)

    async def play_audio_asset(self, asset_name: str) -> None:
        filepath = f"/system/audio/{asset_name}.wav"

        request_id = self._control.get_next_request_id()
        async with self._control.ExecuteFile(request_id, filepath) as (resp, evts):
            async with self._cancellation(request_id=request_id):
                handle_response("ExecuteFile", resp)
                await handle_events("ExecuteFile", evts)

    async def set_led(self, mask: LEDMask, red: float, green: float, blue: float) -> None:
        protocol_mask = conversions.led_to_protocol(mask)
        async with self._control.SetLED(protocol_mask, int(red * 255), int(green * 255), int(blue * 255), 255) as (
            response,
            _,
        ):
            handle_response("SetLED", response)

    async def line_navigation(self, direction: Direction, follow: bool) -> None:
        direction_protocol = conversions.intersection_direction_to_protocol(direction)

        action = Types.LineNavigationAction.Follow if follow else Types.LineNavigationAction.DoNotFollow
        request_id = self._control.get_next_request_id()
        async with self._control.LineNavigation(request_id, direction_protocol, action) as (
            resp,
            evts,
        ):
            async with self._cancellation(request_id=request_id):
                event = await handle_events("LineNavigation", evts)

        intersection = conversions.intersection_bitmap_from_protocol(event.intersection)
        sample = SampleWithoutTimestamp(intersection)
        await self.memory.intersection_queue.write(sample)

    @contextlib.asynccontextmanager
    async def _cancellation(self, *, request_id: int) -> typing.AsyncIterator[None]:
        try:
            yield
        except BaseException as e:
            await _stop_execution(self._control, request_id=request_id)
            raise e



from ozobot.common.exceptions import OzobotError


class EvoError(OzobotError):
    """Base EVO error"""


class OzobotProtocolCommandError(EvoError):
    def __init__(self, command: str, return_value: str, *, description: str | None = None) -> None:
        if description:
            details = f"(returned: {return_value}, description: {description})"
        else:
            details = f"({return_value})"

        super().__init__(f"Protocol command failed: {command} {details}")


class OzobotDataTypeError(EvoError):
    def __init__(self, expected_type: type, received_type: type) -> None:
        super().__init__(f"Unexpected type: expected '{expected_type}' but got '{received_type}'")


class UnsupportedColorCodeNumberError(EvoError):
    def __init__(self, number: int) -> None:
        super().__init__(f"Received color code has unsupported color segment: {number}")
